# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# Taken from:
#  https://github.com/MozillaSecurity/fuzzfetch/blob/1a930d7/tests/conftest.py
import gzip
import logging
import os
from pathlib import Path
from urllib.error import HTTPError
from urllib.request import Request, urlopen

import pytest
import requests_mock

BUILD_CACHE = os.getenv("BUILD_CACHE") == "1"  # set BUILD_CACHE=1 to populate cache
CACHE_PATH = Path(__file__).resolve().parent  # store cache alongside this file
LOG = logging.getLogger("fetcher_cb")
# hosts to cache requests for
# this will create directories like f"mock-{key}"
MOCK_HOSTS = {
    "firefoxci": "https://firefox-ci-tc.services.mozilla.com",
    "hg": "https://hg.mozilla.org",
    "product-details": "https://product-details.mozilla.org",
    "queue": "https://queue.taskcluster.net",
}


assert not BUILD_CACHE or str is not bytes, "BUILD_CACHE requires Python 3"


def _translate_to_path(url):
    assert url.split("://")[0] in {"http", "https"}, "unhandled protocol: %s" % (url,)
    for mock, host in MOCK_HOSTS.items():
        url = url.replace(host, "mock-%s" % (mock,))
    assert url.startswith("mock-"), "unmocked URL: %s" % (url,)
    return CACHE_PATH / url.replace("/", os.sep)


def _cache_requests(request, context):
    """
    request handler for requests.mock
    """
    LOG.debug("%s %r", request.method, request.url)
    path = _translate_to_path(request.url)
    if path.is_file():
        context.status_code = 200
        data = path.read_bytes()
        LOG.debug("-> 200 (%d bytes from %s)", len(data), path)
        return data
    if path.is_dir() and (path / ".get").is_file():
        path = path / ".get"
        context.status_code = 200
        data = path.read_bytes()
        LOG.debug("-> 200 (%d bytes from %s)", len(data), path)
        return data
    # download to cache in mock directories
    if BUILD_CACHE:
        folder = path.parent
        try:
            if not folder.is_dir():
                folder.mkdir(parents=True)
        except OSError:
            # see if any of the leaf folders are actually files
            while folder.resolve() != CACHE_PATH:
                if folder.is_file():
                    # need to rename
                    tmp_folder = folder.parent / (folder.name + ".tmp")
                    folder.rename(tmp_folder)
                    path.parent.mkdir(parents=True)
                    tmp_folder.rename(folder / ".get")
                    break
                folder = folder.parent
        urllib_request = Request(
            request.url,
            request.body if request.method == "POST" else None,
            request.headers,
        )
        try:
            real_http = urlopen(urllib_request)
        except HTTPError as exc:
            context.status_code = exc.code
            return None
        data = real_http.read()
        if data[:2] == b"\x1f\x8b":  # gzip magic number
            data = gzip.decompress(data)  # pylint: disable=no-member
        path.write_bytes(data)
        context.status_code = real_http.getcode()
        LOG.debug("-> %d (%d bytes from http)", context.status_code, len(data))
        return data
    context.status_code = 404
    LOG.debug("-> 404 (at %s)", path)
    return None


@pytest.fixture
def requests_mock_cache():
    """create/use a cache of all requests defined in MOCK_HOSTS

    use env BUILD_CACHE=1 to populate the cache

    the cache is stored in CACHE_PATH/mock-*
    """
    with requests_mock.Mocker() as req_mock:
        req_mock.register_uri(
            requests_mock.ANY, requests_mock.ANY, content=_cache_requests
        )
        yield req_mock



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from .browser import BrowserArgs, BrowserEvaluator
from .js import JSArgs, JSEvaluator
from .base import Evaluator, EvaluatorResult

__all__ = [
    "BrowserArgs",
    "BrowserEvaluator",
    "Evaluator",
    "EvaluatorResult",
    "JSArgs",
    "JSEvaluator",
]



import re
from configparser import NoOptionError, NoSectionError

import pytest

from autobisect import config


def test_config_init_no_config_file(tmp_path):
    """
    Initialize a config object with default settings
    """
    dir_path = tmp_path
    config.CONFIG_DIR = dir_path
    config.CONFIG_FILE = dir_path / "autobisect.ini"

    config.DEFAULT_CONFIG = re.sub(
        r"(?<=storage-path: )(.+)", str(dir_path), config.DEFAULT_CONFIG
    )
    result = config.BisectionConfig()
    assert str(result.db_path) == str(dir_path / "autobisect.db")
    assert result.persist is True
    assert result.persist_limit == 31457280000
    assert str(result.store_path) == str(dir_path)


def test_config_init_with_predefined_config(tmp_path):
    """
    Initialize a config object with predefined settings
    """
    conf = tmp_path / "autobisect.ini"
    conf.write_text(re.sub(r"(?<=persist-limit: )(\d+)", "1000", config.DEFAULT_CONFIG))
    result = config.BisectionConfig(str(conf))
    assert result.persist_limit == 1048576000


def test_config_invalid_config_path(tmp_path):
    """
    Attempts to initialize a config object with a non-existent path
    """
    with pytest.raises(IOError, match="Invalid configuration file specified"):
        config.BisectionConfig(str(tmp_path / "foobar"))


def test_config_init_with_invalid_configuration(tmp_path):
    """
    Attempt to initialize a config object with an invalid configuration
    """
    invalid = tmp_path / "invalid"
    invalid.touch()
    with pytest.raises((NoOptionError, NoSectionError)):
        config.BisectionConfig(str(invalid))


def test_config_init_with_non_existent_dir(tmp_path):
    """
    Attempt to initialize a config object with an invalid configuration
    """
    dir_path = tmp_path / "non-existent"
    config.CONFIG_DIR = dir_path
    config.CONFIG_FILE = dir_path / "autobisect.ini"
    config.BisectionConfig()
    assert dir_path.is_dir()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# flake8: noqa
from .evaluators import *
from . import bisect
from . import build_manager
from . import builds
from . import config
from . import main



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
import argparse
import os

from fuzzfetch import FetcherArgs


def _remove_arg(parser, arg):
    """
    Remove args from existing parser
    :param parser: Target parser
    :param arg: Dest of arg to remove
    """
    # Remove defaults
    if arg in parser._defaults:
        del parser._defaults[arg]

    # Identify actions to be removed
    action_queue = []
    for a in parser._actions:
        if (a.option_strings and a.option_strings[0] == arg) or a.dest == arg:
            action_queue.append(a)

    # Remove actions
    for a in action_queue:
        parser._remove_action(a)

    # Identify action_groups to be removed
    empty_groups = []
    for a in parser._action_groups:
        action_group_queue = []
        for x in a._group_actions:
            if x.dest == arg:
                action_group_queue.append(x)
        for x in action_group_queue:
            a._group_actions.remove(x)
            if len(a._group_actions) == 0 and a not in empty_groups:
                empty_groups.append(a)

    # Remove now emptied groups
    for group in empty_groups:
        parser._action_groups.remove(group)


def _suppress_arg(parser, dest):
    """
    Suppress help messages from existing parser
    :param parser: Target parser
    :param dest: Dest of arg to remove
    """
    for action in parser._actions:
        if action.dest == dest:
            action.help = argparse.SUPPRESS


class ExpandPath(argparse.Action):
    """
    Expand user and relative-paths
    """

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, os.path.abspath(os.path.expanduser(values)))


class BisectCommonArgs(FetcherArgs):
    """
    Arguments common to all bisection targets
    """

    CONFLICTING_ARGS = [
        "build",
        "dry_run",
        "name",
        "nearest",
        "out",
        "target",
        "tests",
        "full_symbols",
    ]

    def __init__(self):
        if not hasattr(self, "parser"):
            self.parser = argparse.ArgumentParser(
                description="Autobisection tool for Mozilla Firefox and SpiderMonkey",
                conflict_handler="resolve",
            )
        super().__init__()

        self.parser.add_argument("testcase", help="Path to testcase")
        boundary_args = self.parser.add_argument_group(
            "Boundary Arguments (YYYY-MM-DD or SHA1 revision)"
        )
        boundary_args.add_argument(
            "--start", help="Start build id (default: earliest available build)"
        )
        boundary_args.add_argument(
            "--end", help="End build id (default: latest available build)"
        )

        bisection_args = self.parser.add_argument_group("Bisection Arguments")
        bisection_args.add_argument(
            "--timeout",
            type=int,
            default=60,
            help="Maximum iteration time in seconds (default: %(default)s)",
        )
        bisection_args.add_argument(
            "--repeat",
            type=int,
            default=1,
            help="Number of times to evaluate testcase (per build)",
        )
        bisection_args.add_argument(
            "--config", action=ExpandPath, help="Path to optional config file"
        )
        bisection_args.add_argument(
            "--find-fix", action="store_true", help="Identify fix date"
        )

        for arg in BisectCommonArgs.CONFLICTING_ARGS:
            _remove_arg(self.parser, arg)

    def sanity_check(self, args):
        """
        Validate supplied args
        :param args: Arguments
        """
        if args.branch is None:
            args.branch = "central"

    def parse_args(self, argv=None):
        """
        Parse and validate supplied args
        :param argv: Arguments
        """
        args = self.parser.parse_args(argv)
        self.sanity_check(args)

        return args


if __name__ == "__main__":
    print(BisectCommonArgs().parse_args())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import logging
import os
import time
from argparse import Action, ArgumentParser
from datetime import timedelta

from fuzzfetch import BuildFlags
from fuzzfetch.fetch import Platform

from .bisect import BisectionResult, Bisector
from .evaluators import BrowserArgs, BrowserEvaluator, JSArgs, JSEvaluator

LOG = logging.getLogger("autobisect")


class ExpandPath(Action):
    """
    Expand user and relative-paths
    """

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, os.path.abspath(os.path.expanduser(values)))


def console_init_logging():
    """
    Enable logging when called from console
    """
    log_level = logging.INFO
    log_fmt = "[%(asctime)s] %(message)s"
    if bool(os.getenv("DEBUG")):
        log_level = logging.DEBUG
        log_fmt = "%(levelname).1s %(name)s [%(asctime)s] %(message)s"
    logging.basicConfig(format=log_fmt, datefmt="%Y-%m-%d %H:%M:%S", level=log_level)
    logging.getLogger("requests").setLevel(logging.WARNING)


def parse_args():
    """
    Argument parser
    """
    parser = ArgumentParser(description="Firefox and Spidermonkey Bisection Tool")
    subparsers = parser.add_subparsers(dest="target")
    subparsers.required = True

    ff_sub = BrowserArgs(subparsers.add_parser("firefox", conflict_handler="resolve"))
    js_sub = JSArgs(subparsers.add_parser("js", conflict_handler="resolve"))

    args = parser.parse_args()
    if args.target == "firefox":
        ff_sub.sanity_check(args)
    elif args.target == "js":
        js_sub.sanity_check(args)

    return args


def main(args):
    """
    Autobisect main entry point
    """
    if args.target == "firefox":
        evaluator = BrowserEvaluator(**vars(args))
    else:
        if args.flags is None:
            args.flags = []
        else:
            args.flags = args.flags.split(" ")

        evaluator = JSEvaluator(**vars(args))

    flags = BuildFlags(
        args.asan, args.tsan, args.debug, args.fuzzing, args.coverage, args.valgrind
    )
    platform = Platform(args.os, args.cpu)
    bisector = Bisector(
        evaluator,
        args.target,
        args.branch,
        args.start,
        args.end,
        flags,
        platform,
        args.find_fix,
        args.config,
    )
    start_time = time.time()
    result = bisector.bisect()
    end_time = time.time()
    if result.status == BisectionResult.SUCCESS:
        LOG.info("Reduced build range to:")
        LOG.info("> Start: %s (%s)", result.start.changeset, result.start.id)
        LOG.info("> End: %s (%s)", result.end.changeset, result.end.id)
        LOG.info("> %s", result.pushlog)
    else:
        LOG.error("Bisection failed!")

    elapsed = timedelta(seconds=(int(end_time - start_time)))
    LOG.info("Bisection completed in: %s", elapsed)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import logging
import os
import shutil
import sqlite3
import time
from contextlib import contextmanager
from pathlib import Path

from .config import BisectionConfig

LOG = logging.getLogger("build-manager")


class DatabaseManager(object):
    """
    Sqlite3 wrapper class
    """

    def __init__(self, db_path):
        self.con = None
        self.cur = None
        self.open(str(db_path))

    def open(self, db_path):
        """
        Opens and initializes the sqlite3 database
        :param db_path: Path to the sqlite3 database
        :type db_path: str
        """
        self.con = sqlite3.connect(db_path)
        self.cur = self.con.cursor()
        self.cur.execute("CREATE TABLE IF NOT EXISTS in_use (build_path, pid INT)")
        self.cur.execute(
            "CREATE TABLE IF NOT EXISTS download_queue (build_path TEXT primary key, pid INT)"
        )

    def close(self):
        """
        Closes the sqlite3 database
        """
        if self.con:
            self.con.commit()
            self.con.close()

    def __del__(self):
        self.close()


class BuildManager(object):
    """
    A class for managing downloaded builds
    """

    def __init__(self, config=None):
        self.config = config or BisectionConfig()
        self.build_dir = self.config.store_path / "builds"
        if not Path.is_dir(self.build_dir):
            os.makedirs(self.build_dir)

        self.pid = os.getpid()
        self.db = DatabaseManager(self.config.db_path)

    @property
    def current_build_size(self):
        """
        Recursively enumerate the size of the supplied build
        """
        return sum(
            [os.path.getsize(f) for f in self.build_dir.rglob("*") if os.path.isfile(f)]
        )

    def enumerate_builds(self):
        """
        Enumerate all available builds including their size and stats
        """
        builds = [os.fspath(x) for x in self.build_dir.iterdir() if x.is_dir()]
        return sorted(builds, key=lambda b: os.stat(b).st_atime)

    def remove_old_builds(self):
        """
        Removes stored builds to make room for newer builds
        """
        while self.current_build_size > self.config.persist_limit:
            builds = self.enumerate_builds()
            for build_path in builds:
                if self.current_build_size < self.config.persist_limit:
                    break

                res = self.db.cur.execute(
                    "SELECT * FROM in_use, download_queue "
                    "WHERE in_use.build_path = ? OR download_queue.build_path = ?",
                    (build_path, build_path),
                )
                if res.fetchone() is None:
                    LOG.debug("Removing build: %s", build_path)
                    shutil.rmtree(build_path)
                self.db.con.commit()

            time.sleep(0.1)

    @contextmanager
    def get_build(self, build):
        """
        Retrieve the build matching the supplied revision
        :param build: A fuzzFetch.Fetcher build object
        """
        # pylint: disable=protected-access
        target = "js" if build._memo["_target"] == "js" else "ff"
        branch = "m-%s" % build._branch[0]
        platform = build._platform.system
        flags = build._flags.build_string()
        rev = build.changeset[:12]
        build_name = "%s-%s-%s%s-%s" % (target, branch, platform, flags, rev)
        target_path = self.build_dir / build_name.lower()
        path_string = os.fspath(target_path)

        try:
            # Insert build_path into in_use to prevent deletion
            self.db.cur.execute(
                "INSERT INTO in_use VALUES (?, ?)", (path_string, self.pid)
            )
            self.db.con.commit()

            # Try to insert the build_path into download_queue
            # If the insert fails, another process is already downloading it
            # Poll the database until it complete
            try:
                self.db.cur.execute(
                    "INSERT into download_queue VALUES (?, ?)", (path_string, self.pid)
                )
                self.db.con.commit()
                # If the build doesn't exist on disk, download it
                if not Path.is_dir(target_path):
                    self.remove_old_builds()
                    build.extract_build(target_path)
            except sqlite3.IntegrityError:
                LOG.warning(
                    "Another process is attempting to download the build. Waiting"
                )
                while True:
                    res = self.db.cur.execute(
                        "SELECT * FROM download_queue WHERE build_path = ?",
                        (path_string,),
                    )
                    if res.fetchone() is None:
                        break

                    time.sleep(0.1)
            finally:
                self.db.cur.execute(
                    "DELETE FROM download_queue WHERE build_path = ? AND pid = ?",
                    (path_string, self.pid),
                )
                self.db.con.commit()

            yield target_path
        finally:
            self.db.cur.execute(
                "DELETE FROM in_use WHERE build_path = ? AND pid = ?",
                (path_string, self.pid),
            )
            self.db.con.commit()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import logging
import os
import re
from string import Template

import requests
from lithium import interestingness

from ..base import Evaluator, EvaluatorResult

LOG = logging.getLogger("js-eval")

HTTP_SESSION = requests.Session()

FLAGS_URL = Template(
    "https://hg.mozilla.org/mozilla-unified/raw-file/$rev/js/src/shell/js.cpp"
)


def _get_rev(binary):
    """
    Return either the revision specified in the fuzzmanagerconf
    :param binary: Path to build
    """
    rev = "tip"
    path = "%s.fuzzmanagerconf" % (os.path.splitext(binary)[0],)
    if os.path.isfile(path):
        with open(path) as in_fp:
            for line in in_fp:
                if line.startswith("product_version"):
                    rev = line.split("-")[1].strip()

    return rev


class JSEvaluatorException(Exception):
    """
    Raised for any JSEvaluator exception
    """


class JSEvaluator(Evaluator):
    """
    Testcase evaluator for SpiderMonkey shells
    """

    def __init__(self, testcase, **kwargs):
        self.testcase = testcase
        self.flags = kwargs.get("flags", [])
        self.repeat = kwargs.get("repeat", 1)

        # JS Shell launch arguments
        self.timeout = kwargs.get("timeout", 60)
        self.detect = kwargs.get("detect", "crash")

        if self.detect == "diff":
            if not kwargs.get("arg_1") or not kwargs.get("arg_2"):
                raise JSEvaluatorException("Diff mode requires 'arg_1' and 'arg_2'")
            self._arg_1 = kwargs.get("arg_1")
            self._arg_2 = kwargs.get("arg_2")
        elif self.detect == "match":
            if not kwargs.get("match"):
                raise JSEvaluatorException("Match mode requires a match string")
            self._match = kwargs.get("match")
            self._regex = kwargs.get("regex")

    @staticmethod
    def get_valid_flags(rev):
        """
        Extract list of runtime flags available to the current build
        :param rev:
        """
        # Fuzzing safe is always included but not in the same format as the rest
        flags = []

        try:
            data = HTTP_SESSION.get(FLAGS_URL.substitute(rev=rev), stream=True)
            data.raise_for_status()
            matches = re.findall(r"(?:get\w+Option)\(\"(.[^\"]*)", data.text)
            flags.extend(list(set(matches)))
        except requests.exceptions.RequestException as e:
            LOG.warning("Failed to retrieve build flags %s", e)

        return flags

    def verify_build(self, binary, flags):
        """
        Verify that build doesn't crash on start
        :param binary: The path to the target binary
        :param flags: Runtime flags
        :return: Boolean
        """
        LOG.info("> Verifying build...")
        args = [binary, *flags, "-e", '"quit()"']
        run_data = interestingness.timed_run.timed_run(args, self.timeout, None)
        if run_data.sta is not interestingness.timed_run.NORMAL:
            LOG.error(">> Build crashed!")
            return False

        return True

    def evaluate_testcase(self, build_path):
        """
        Validate build and launch with supplied testcase
        :return: Result of evaluation
        """
        binary = os.path.join(build_path, "dist", "bin", "js")
        rev = _get_rev(binary)
        all_flags = self.get_valid_flags(rev)
        flags = []
        for flag in self.flags:
            if flag.lstrip("--").split("=")[0] in all_flags:
                flags.append(flag)

        if os.path.isfile(binary) and self.verify_build(binary, flags):
            common_args = ["-t", "%s" % self.timeout, binary, *flags, self.testcase]

            try:
                # Some testcases require setting the cwd to the parent dir
                previous_path = os.getcwd()
                os.chdir(os.path.dirname(self.testcase))

                for _ in range(self.repeat):
                    LOG.info("> Launching build with testcase...")
                    if self.detect == "diff":
                        args = ["-a", self._arg_1, "-b", self._arg_2] + common_args
                        if interestingness.diff_test.interesting(args, None):
                            return EvaluatorResult.BUILD_CRASHED
                    elif self.detect == "output":
                        args = [self._match] + common_args
                        if interestingness.outputs.interesting(args, None):
                            return EvaluatorResult.BUILD_CRASHED
                    elif self.detect == "crash":
                        if interestingness.crashes.interesting(common_args, None):
                            return EvaluatorResult.BUILD_CRASHED
                    elif self.detect == "hang":
                        if interestingness.hangs.interesting(common_args, None):
                            return EvaluatorResult.BUILD_CRASHED
            finally:
                # Reset cwd
                os.chdir(previous_path)

            return EvaluatorResult.BUILD_PASSED

        return EvaluatorResult.BUILD_FAILED



import random
from datetime import datetime, timedelta

from autobisect.builds import BuildRange


def test_build_range_basic_operation():
    """
    Simple test
    """
    builds = BuildRange([1, 2, 3])
    assert len(builds) == 3
    assert builds.builds == [1, 2, 3]


def test_build_range_build_access_by_index():
    """
    Simple test to ensure build_range's __getitem__ returns the builds build_info property
    """
    builds = ["abc", "zyx", "foo", "bar"]
    build_range = BuildRange(builds)
    for index, key in enumerate(builds, start=0):
        assert build_range[index] == key


def test_build_range_splice():
    """
    Test splicing of BuildRange
    """
    builds = ["abc", "zyx", "foo", "bar"]
    build_range = BuildRange(builds)
    copy = build_range[:2]
    assert len(copy) == 2
    assert copy[0] == "abc"
    assert copy[1] == "zyx"


def test_build_range_mid_point():
    """
    Test to ensure proper midpoint is returned
    """
    for i in range(10):
        mid_point = i // 2 if i > 0 else None
        assert BuildRange(list(range(i))).mid_point == mid_point


def test_build_range_random():
    """
    Test to ensure random returns element in list
    """
    builds = list(range(10))
    assert BuildRange(builds).random in builds


def test_build_range_random_empty_list():
    """
    Test BuildRange.random with an empty list
    """
    assert BuildRange([]).random is None


def test_build_range_indexing():
    """
    Test to ensure proper index is returned
    """
    builds = ["abc", "zyx", "foo", "bar"]
    build_range = BuildRange(builds)
    for k1, k2 in zip(build_range, builds):
        assert k1 == k2
    for index, build in enumerate(builds, start=0):
        assert build_range.index(build) == index


def test_build_range_new_classmethod():
    """
    Simple test to ensure build range is expected length using new method
    """
    days = random.randint(1, 100)
    build_range = BuildRange.new(datetime.now(), datetime.now() + timedelta(days=days))
    assert len(build_range) == days + 1



import pytest

from autobisect.evaluators import BrowserEvaluator
from autobisect.evaluators import EvaluatorResult
from autobisect.evaluators.browser.browser import BrowserEvaluatorException


def test_verify_build_status(mocker):
    """
    Test that BrowserEvaluator.verify_build() returns the correct status
    """
    mocker.patch(
        "autobisect.BrowserEvaluator.launch",
        side_effect=(
            EvaluatorResult.BUILD_CRASHED,
            EvaluatorResult.BUILD_PASSED,
            EvaluatorResult.BUILD_FAILED,
        ),
    )
    browser = BrowserEvaluator("testcase.html")
    assert browser.verify_build(None) is False
    assert browser.verify_build(None) is True
    assert browser.verify_build(None) is False


def test_evaluate_testcase_simple(mocker, tmp_path):
    """
    Simple test of BrowserEvaluator.evaluate_testcase()
    """
    mocker.patch(
        "autobisect.BrowserEvaluator.launch",
        autospec=True,
        return_value=EvaluatorResult.BUILD_PASSED,
    )
    browser = BrowserEvaluator("testcase.html")
    (tmp_path / "firefox").touch()
    assert browser.evaluate_testcase(tmp_path) == EvaluatorResult.BUILD_PASSED


def test_evaluate_testcase_non_existent_binary(tmp_path):
    """
    Test that BrowserEvaluator.evaluate_testcase fails when using a non-existent build path
    """
    browser = BrowserEvaluator("testcase.html")
    assert browser.evaluate_testcase(tmp_path) == EvaluatorResult.BUILD_FAILED


def test_launch_simple(mocker, tmp_path):
    """
    Simple test of BrowserEvaluator.launch()
    """
    mocker.patch("grizzly.replay.ReplayManager.main", side_effect=(False, True))

    binary = tmp_path / "firefox"
    binary.touch()
    testcase = tmp_path / "testcase.html"
    testcase.touch()

    browser = BrowserEvaluator(testcase)
    assert browser.launch(binary, testcase, 1) == EvaluatorResult.BUILD_CRASHED
    assert browser.launch(binary, testcase, 1) == EvaluatorResult.BUILD_PASSED


def test_launch_non_existent_binary(tmp_path):
    """
    Simple test of BrowserEvaluator.launch()
    """
    binary = tmp_path / "firefox"
    testcase = tmp_path / "testcase.html"
    testcase.touch()

    browser = BrowserEvaluator(testcase)

    with pytest.raises(BrowserEvaluatorException):
        browser.launch(binary, testcase, 1)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
import logging
import random
from datetime import datetime, timedelta
from enum import Enum

import requests
from fuzzfetch import BuildFlags, Fetcher, FetcherException

from .build_manager import BuildManager
from .builds import BuildRange
from .evaluators import EvaluatorResult

LOG = logging.getLogger("bisect")


def get_autoland_range(start, end):
    """
    Retrieve changeset from autoland within supplied boundary

    :param start: Starting revision
    :param end: Ending revision
    """
    url = (
        "https://hg.mozilla.org/mozilla-central/json-pushes"
        f"?fromchange={start}&tochange={end}"
    )
    try:
        data = requests.get(url)
        data.raise_for_status()
    except requests.exceptions.RequestException as exc:
        LOG.error("Failed to retrieve autoland changeset %s", exc)
        return None

    json = data.json()
    key_len = len(json.keys())
    if key_len == 1:
        push_id = list(json.keys())[0]
        return json[push_id]["changesets"]

    LOG.warning(f"Detected {key_len} top-level changes.  Cannot bisect into autoland.")
    return None


class StatusException(Exception):
    """
    Raised when an invalid status is supplied
    """


class VerificationStatus(Enum):
    """
    Class for storing build verification result
    """

    SUCCESS = 0
    START_BUILD_FAILED = 1
    END_BUILD_FAILED = 2
    START_BUILD_CRASHES = 3
    END_BUILD_PASSES = 4
    FIND_FIX_START_BUILD_PASSES = 5
    FIND_FIX_END_BUILD_CRASHES = 6

    @property
    def message(self):
        """
        Return message matching explaining current status
        """
        result = None
        if self == self.SUCCESS:
            result = "Verified supplied boundaries!"
        elif self == self.START_BUILD_FAILED:
            result = "Unable to launch the start build!"
        elif self == self.END_BUILD_FAILED:
            result = "Unable to launch the end build!"
        elif self == self.START_BUILD_CRASHES:
            result = "Testcase reproduces on start build!"
        elif self == self.END_BUILD_PASSES:
            result = "Testcase does not reproduce on end build!"
        elif self == self.FIND_FIX_START_BUILD_PASSES:
            result = "Start build didn't crash!"
        elif self == self.FIND_FIX_END_BUILD_CRASHES:
            result = "End build crashes!"

        return result


class BisectionResult(object):
    """
    Class for storing bisection result
    """

    SUCCESS = 0
    FAILED = 1

    def __init__(self, status, start, end, branch, message=None):
        self.status = status
        self.start = start
        self.end = end
        self.branch = branch
        if status == BisectionResult.SUCCESS:
            if start.build_info["moz_source_repo"] == end.build_info["moz_source_repo"]:
                base = start.build_info["moz_source_repo"]
            else:
                base = "https://hg.mozilla.org/mozilla-unified"

            self.pushlog = (
                f"{base}/pushloghtml?fromchange="
                f"{start.changeset}&tochange={end.changeset}"
            )

        self.message = message


class Bisector(object):
    """
    Taskcluster Bisection Class
    """

    def __init__(
        self,
        evaluator,
        target,
        branch,
        start,
        end,
        flags,
        platform,
        find_fix=False,
        config=None,
    ):
        """
        Instantiate bisection object
        :param evaluator: Object instance used to evaluate testcase
        :param target: Type of builds to fetch
        :param branch: Mozilla branch to use for finding builds
        :param start: Start revision, date, or buildid
        :param end: End revision, date, or buildid
        :param flags: Build flags (asan, tsan, debug, fuzzing, valgrind)
        :param platform: fuzzfetch.fetch.Platform instance
        :param find_fix: Boolean identifying whether to find a fix or bisect bug
        :param config: Path to config file
        """
        self.evaluator = evaluator
        self.target = target
        self.branch = (
            Fetcher.resolve_esr(branch) if branch.startswith("esr") else branch
        )
        self.platform = platform
        self.flags = BuildFlags(*flags)
        self.find_fix = find_fix

        # If no start date is supplied, default to oldest available build
        earliest = (datetime.utcnow() - timedelta(days=364)).strftime("%Y-%m-%d")
        start_id = start if start else earliest
        end_id = end if end else "latest"

        self.start = Fetcher(
            self.target,
            self.branch,
            start_id,
            self.flags,
            self.platform,
            Fetcher.BUILD_ORDER_ASC,
        )
        self.end = Fetcher(
            self.target,
            self.branch,
            end_id,
            self.flags,
            self.platform,
            Fetcher.BUILD_ORDER_DESC,
        )

        self.build_manager = BuildManager(config)

    def _get_daily_builds(self):
        """
        Create build range containing one build per day
        """
        start = self.start.datetime + timedelta(days=1)
        end = self.end.datetime - timedelta(days=1)
        LOG.info(f"Enumerating daily builds: {start} - {end}")

        return BuildRange.new(start, end)

    def _get_pushdate_builds(self):
        """
        Create build range containing all builds per pushdate
        """
        start = self.start.datetime
        end = self.end.datetime
        LOG.info(f"Enumerating pushdate builds: {start} - {end}")

        builds = []
        for dt in [start, end]:
            date = dt.strftime("%Y-%m-%d")
            for build in Fetcher.iterall(self.target, self.branch, date, self.flags):
                # Only keep builds after the start and before the end boundaries
                if self.end.datetime > build.datetime > self.start.datetime:
                    builds.append(build)

        return BuildRange(builds)

    def _get_autoland_builds(self):
        """
        Create build range containing all autoland builds per pushdate
        """
        if self.branch != "central":
            return []

        start = self.start.datetime
        end = self.end.datetime

        LOG.info(f"Enumerating autoland builds: {start} - {end}")
        changesets = get_autoland_range(self.start.changeset, self.end.changeset)
        if changesets is None:
            return []

        builds = []
        for changeset in changesets:
            try:
                build = Fetcher(
                    self.target, "autoland", changeset, self.flags, self.platform
                )
                builds.append(build)
            except FetcherException:
                LOG.warning("Unable to find build for %s", changeset)

        return BuildRange(builds)

    def build_iterator(self, build_range, random_choice):
        """
        Yields next build to be evaluated until all possibilities consumed
        """
        while build_range:
            if random_choice:
                build = random.choice(build_range)
            else:
                build = build_range.mid_point

            index = build_range.index(build)
            if not isinstance(build, Fetcher):
                try:
                    build = Fetcher(
                        self.target, self.branch, build, self.flags, self.platform
                    )
                except FetcherException:
                    LOG.warning("Unable to find build for %s", build)
                    build_range.builds.remove(build)
                    continue

            status = yield build
            build_range = self.update_range(status, build, index, build_range)

        yield None

    def bisect(self, random_choice=False):
        """
        Main bisection function

        :param random_choice: Select builds at random during bisection (QuickSort)

        :return: BisectionResult
        """
        LOG.info("Begin bisection...")
        LOG.info("> Start: %s (%s)", self.start.changeset, self.start.id)
        LOG.info("> End: %s (%s)", self.end.changeset, self.end.id)

        verified = self.verify_bounds()
        if verified == VerificationStatus.SUCCESS:
            LOG.info(verified.message)
        else:
            LOG.critical(verified.message)
            return BisectionResult(
                BisectionResult.FAILED,
                self.start,
                self.end,
                self.branch,
                verified.message,
            )

        LOG.info("Attempting to reduce bisection range using taskcluster binaries")
        strategies = [
            self._get_daily_builds,
            self._get_pushdate_builds,
            self._get_autoland_builds,
        ]
        for strategy in strategies:
            build_range = strategy()
            generator = self.build_iterator(build_range, random_choice)
            next_build = next(generator)
            while next_build is not None:
                status = self.test_build(next_build)
                next_build = generator.send(status)

        return BisectionResult(
            BisectionResult.SUCCESS, self.start, self.end, self.branch
        )

    def update_range(self, status, build, index, build_range):
        """
        Returns a new build range based on the status of the previously evaluated test

        :param status: The status of the evaluated testcase
        :param build: The evaluated build
        :param index: Index of the build
        :param build_range: The build_range to update
        :return: The adjusted BuildRange object
        """
        if status == EvaluatorResult.BUILD_PASSED:
            if not self.find_fix:
                self.start = build
                return build_range[index + 1 :]

            self.end = build
            return build_range[:index]
        if status == EvaluatorResult.BUILD_CRASHED:
            if not self.find_fix:
                self.end = build
                return build_range[:index]

            self.start = build
            return build_range[index + 1 :]
        if status == EvaluatorResult.BUILD_FAILED:
            range_copy = build_range[:]
            range_copy.builds.pop(index)
            return range_copy

        raise StatusException("Invalid status supplied")

    def test_build(self, build):
        """
        Prepare the build directory and launch the supplied build
        :param build: An Fetcher object to prevent duplicate fetching
        :return: The result of the build evaluation
        """
        LOG.info("Testing build %s (%s)", build.changeset, build.id)
        # If persistence is enabled and a build exists, use it
        with self.build_manager.get_build(build) as build_path:
            return self.evaluator.evaluate_testcase(build_path)

    def verify_bounds(self):
        """
        Verify that the supplied bounds behave as expected
        :return: Boolean
        """
        LOG.info("Attempting to verify boundaries...")
        start_result = self.test_build(self.start)
        if start_result not in set(EvaluatorResult):
            raise StatusException("Invalid status supplied")

        if start_result == EvaluatorResult.BUILD_FAILED:
            return VerificationStatus.START_BUILD_FAILED
        if start_result == EvaluatorResult.BUILD_CRASHED and not self.find_fix:
            return VerificationStatus.START_BUILD_CRASHES
        if start_result == EvaluatorResult.BUILD_PASSED and self.find_fix:
            return VerificationStatus.FIND_FIX_START_BUILD_PASSES

        end_result = self.test_build(self.end)
        if end_result not in set(EvaluatorResult):
            raise StatusException("Invalid status supplied")

        if end_result == EvaluatorResult.BUILD_FAILED:
            return VerificationStatus.END_BUILD_FAILED
        if end_result == EvaluatorResult.BUILD_PASSED and not self.find_fix:
            return VerificationStatus.END_BUILD_PASSES
        if end_result == EvaluatorResult.BUILD_CRASHED and self.find_fix:
            return VerificationStatus.FIND_FIX_END_BUILD_CRASHES

        return VerificationStatus.SUCCESS



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from ...args import BisectCommonArgs


class JSArgs(BisectCommonArgs):
    """
    Argparser for JSEvaluator
    """

    def __init__(self, parser):
        self.parser = parser
        super().__init__()

        launcher = self.parser.add_argument_group("launcher arguments")
        launcher.add_argument("--flags", help="Runtime flags to pass to the binary")
        launcher.add_argument(
            "--detect",
            choices=["crash", "diff", "hang", "output"],
            default="crash",
            help="Type of failure to detect (default: %(default)s)",
        )

        diff = self.parser.add_argument_group("diff arguments")
        diff.add_argument("--arg_1", help="Arguments to supply to the first run")
        diff.add_argument("--arg_2", help="Arguments to supply to the second run")

        output = self.parser.add_argument_group("output arguments")
        output.add_argument("--match", help="String to detect in output")
        output.add_argument("--regex", help="Treat match as a regex")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import configparser
import logging
from pathlib import Path

LOG = logging.getLogger("browser-bisect")

CONFIG_DIR = Path.home() / ".autobisect"
CONFIG_FILE = CONFIG_DIR / "autobisect.ini"

DEFAULT_CONFIG = (
    """
[autobisect]
storage-path: %s
persist: true
; size in MBs
persist-limit: 30000
"""
    % CONFIG_DIR
)


class BisectionConfig(object):
    """
    Class for accessing configuration data and 'skip' revs
    """

    def __init__(self, config_file=None):
        """
        Initializes the object using either the specified config_file or creates a new
        database using default values
        :param config_file: A path to custom configuration file
        :type config_file: str
        """

        if not config_file:
            config_file = self.create_default_config()
        elif not Path(config_file).is_file():
            raise IOError("Invalid configuration file specified")

        config_obj = configparser.ConfigParser()
        config_obj.read(config_file)

        try:
            self.persist = config_obj.getboolean("autobisect", "persist")
            persist_limit = (
                config_obj.getint("autobisect", "persist-limit") * 1024 * 1024
            )
            self.persist_limit = persist_limit if self.persist else 0
            self.store_path = Path(config_obj.get("autobisect", "storage-path"))
        except (configparser.NoOptionError, configparser.NoSectionError) as e:
            LOG.critical("Unable to parse configuration file: %s", e.message)
            raise

        self.db_path = self.store_path / "autobisect.db"

    @staticmethod
    def create_default_config():
        """
        Create a config file using default options and write to disk
        :return: A path to the newly created configuration file
        :rtype: str
        """
        if not CONFIG_DIR.is_dir():
            CONFIG_DIR.mkdir(parents=True)
        if not CONFIG_FILE.is_file():
            CONFIG_FILE.write_text(DEFAULT_CONFIG)

        return str(CONFIG_FILE)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from abc import ABC, abstractmethod
from enum import Enum


class EvaluatorResult(Enum):
    """
    Evaluator result
    """

    BUILD_CRASHED = 0
    BUILD_PASSED = 1
    BUILD_FAILED = 2


class Evaluator(ABC):
    """
    Base evaluator class
    """

    @abstractmethod
    def evaluate_testcase(self, build_path):
        """
        Method for evaluating testcase

        :returns: Evaluation result
        :rtype: EvaluatorResult
        """



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Code originally taken from mozregression
# https://github.com/mozilla/mozregression/blob/5b986a3165a5208dd0722d6fc882b47e7fc1b627/mozregression/build_range.py

import copy
import logging
import random
from datetime import timedelta

LOG = logging.getLogger("builds")


class BuildRange(object):
    """
    A class for storing a range of builds or build representations
    """

    def __init__(self, builds):
        """
        Instantiate a new instance

        param builds: A list of Fetcher.BuildTask objects
        """
        self._builds = builds

    def __len__(self):
        return len(self._builds)

    def __getitem__(self, expr):
        if isinstance(expr, slice):
            new_range = copy.copy(self)
            new_range._builds = self._builds[expr]
            return new_range

        return self._builds[expr]

    @property
    def builds(self):
        """
        Returns the "builds" list
        """
        return self._builds

    @property
    def mid_point(self):
        """
        Returns the midpoint of the "builds" list
        """
        if len(self) > 0:
            return self._builds[len(self) // 2]

        return None

    @property
    def random(self):
        """
        Select a random index
        """
        if len(self) > 0:
            return random.choice(self._builds)

        return None

    def index(self, build):
        """
        Returns the index of the provided build
        :param build: An object within the "builds" list
        """
        return self.builds.index(build)

    @classmethod
    def new(cls, start, end):
        """
        Creates a list of builds between two ranges
        :param start: A starting datetime object
        :type start: datetime.datetime
        :param end: An ending datetime object
        :type end: datetime.datetime
        :return: A BuildRange object
        """
        dates = []
        # Remove time date
        start = start.replace(hour=0, minute=0, second=0, microsecond=0)
        end = end.replace(hour=0, minute=0, second=0, microsecond=0)
        delta = end - start
        for offset in range(delta.days + 1):
            dates.append((start + timedelta(days=offset)).strftime("%Y-%m-%d"))

        return cls(dates)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autobisect',
 'autobisect.evaluators',
 'autobisect.evaluators.browser',
 'autobisect.evaluators.browser.tests',
 'autobisect.evaluators.js',
 'autobisect.tests']

package_data = \
{'': ['*'],
 'autobisect.tests': ['mock-firefoxci/api/index/v1/task/*',
                      'mock-firefoxci/api/queue/v1/task/BfFeMY14Qyu_rMA2pxfZZg/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/BfFeMY14Qyu_rMA2pxfZZg/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/DYaFBxzITgCwBUp48Bugtw/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/DYaFBxzITgCwBUp48Bugtw/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/EYhOnvRFQd-rMlW2oAl10A/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/EYhOnvRFQd-rMlW2oAl10A/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/FnPonlT1QVSw-B0M0KcotQ/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/FnPonlT1QVSw-B0M0KcotQ/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/H5Dl5ijpSpOPwMLEEWx_bA/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/H5Dl5ijpSpOPwMLEEWx_bA/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/HFsFsbogQ0O-sFZoYDmAZw/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/HFsFsbogQ0O-sFZoYDmAZw/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/HNaxi6w_RQeXGb4fYQFtwA/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/HNaxi6w_RQeXGb4fYQFtwA/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/HPE2LlTUQU62unXX8GvNGA/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/HPE2LlTUQU62unXX8GvNGA/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/HX3oSg_SRTq5-Mc5D0PPHg/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/HX3oSg_SRTq5-Mc5D0PPHg/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/Hq2DMUk6QP26axMV8fDi9w/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/Hq2DMUk6QP26axMV8fDi9w/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/JucwusRxQHKQwDUPoOJIhQ/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/JucwusRxQHKQwDUPoOJIhQ/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/NKFp2FPpQ06E8C7N30JPdw/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/NKFp2FPpQ06E8C7N30JPdw/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/O1Sh28LhQC62SfLAi0orxg/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/O1Sh28LhQC62SfLAi0orxg/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/R2zRDt-ATf-zE-hYXR8kEw/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/R2zRDt-ATf-zE-hYXR8kEw/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/SDAzDRggRPOPYHEQC4WDEw/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/SDAzDRggRPOPYHEQC4WDEw/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/Tp5LesydQWyPNqpMThZLyg/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/Tp5LesydQWyPNqpMThZLyg/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/TxMd0rIfQKucR6p0gkaMSw/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/TxMd0rIfQKucR6p0gkaMSw/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/UXVT--1qQfeaGtNICDKx3w/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/UXVT--1qQfeaGtNICDKx3w/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/WLcjhVjtTLW0A_ued1Io-Q/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/WLcjhVjtTLW0A_ued1Io-Q/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/Zb_1byPDSTyhUkZeCg-PZg/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/Zb_1byPDSTyhUkZeCg-PZg/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/cRT3ROAzReG2wv7PW08JMg/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/cRT3ROAzReG2wv7PW08JMg/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/cSP0vuVWTX6JspZ2VTtDdg/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/cSP0vuVWTX6JspZ2VTtDdg/artifacts/public/build/*',
                      'mock-firefoxci/api/queue/v1/task/fXYHDLC_RkWNtOv01aB8wQ/artifacts/*',
                      'mock-firefoxci/api/queue/v1/task/fXYHDLC_RkWNtOv01aB8wQ/artifacts/public/build/*',
                      'mock-hg/mozilla-central/*']}

install_requires = \
['black>=19.10b0,<20.0',
 'fuzzfetch>=1.0.4,<2.0.0',
 'grizzly-framework>=0.11.1,<0.12.0',
 'lithium-reducer>=0.3.3,<0.4.0',
 'toml>=0.9,<0.10']

entry_points = \
{'console_scripts': ['autobisect = autobisect.main:main']}

setup_kwargs = {
    'name': 'autobisect',
    'version': '0.8.4',
    'description': 'Automatic bisection utility for Mozilla Firefox and SpiderMonkey',
    'long_description': "Autobisect\n==========\n[![Build Status](https://travis-ci.com/MozillaSecurity/autobisect.svg?branch=master)](https://travis-ci.org/MozillaSecurity/autobisect)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![codecov](https://codecov.io/gh/MozillaSecurity/autobisect/branch/master/graph/badge.svg)](https://codecov.io/gh/MozillaSecurity/autobisect)\n\nAutobisect is a python module that automates bisection of Mozilla Firefox and SpiderMonkey bugs.\n\nInstallation\n------------\n\n```bash\ngit clone git@github.com:MozillaSecurity/autobisect.git\ncd autobisect\npoetry install\n```\n\nUsage\n-----\nFirefox bug bisection supports the following arguments:\n\n```\npython -m autobisect firefox --help\n\npositional arguments:\n  testcase              Path to testcase\n\noptional arguments:\n  -h, --help            show this help message and exit\n\nLauncher Arguments:\n  -e EXTENSION, --extension EXTENSION\n                        Install an extension. Specify the path to the xpi or\n                        the directory containing the unpacked extension. To\n                        install multiple extensions specify multiple times\n  --launch-timeout LAUNCH_TIMEOUT\n                        Number of seconds to wait before LaunchError is raised\n                        (default: 300)\n  --log-limit LOG_LIMIT\n                        Log file size limit in MBs (default: 'no limit')\n  -m MEMORY, --memory MEMORY\n                        Browser process memory limit in MBs (default: 'no\n                        limit')\n  -p PREFS, --prefs PREFS\n                        prefs.js file to use\n  --soft-asserts        Detect soft assertions\n  --xvfb                Use Xvfb (Linux only)\n\nReporter Arguments:\n  --ignore IGNORE [IGNORE ...]\n                        Space separated ignore list. ie: log-limit memory\n                        timeout (default: nothing)\n\nReplay Arguments:\n  --any-crash           Any crash is interesting, not only crashes which match\n                        the original signature.\n  --idle-timeout IDLE_TIMEOUT\n                        Number of seconds to wait before polling testcase for\n                        idle (default: 60)\n  --idle-threshold IDLE_THRESHOLD\n                        CPU usage threshold to mark the process as idle\n                        (default: 25)\n\nTarget:\n  --target {firefox,js}\n                        Specify the build target. (default: firefox)\n  --os {Android,Darwin,Linux,Windows}\n                        Specify the target system. (default: Linux)\n  --cpu {AMD64,ARM64,aarch64,arm,arm64,i686,x64,x86,x86_64}\n                        Specify the target CPU. (default: x86_64)\n\nBuild:\n  --build DATE|REV|NS   Specify the build to download, (default: latest)\n                        Accepts values in format YYYY-MM-DD (2017-01-01)\n                        revision (57b37213d81150642f5139764e7044b07b9dccc3) or\n                        TaskCluster namespace (gecko.v2....)\n\nBranch:\n  --inbound             Download from mozilla-inbound\n  --central             Download from mozilla-central (default)\n  --release             Download from mozilla-release\n  --beta                Download from mozilla-beta\n  --esr-stable          Download from esr-stable\n  --esr-next            Download from esr-next\n  --try                 Download from try\n\nBuild Arguments:\n  -d, --debug           Get debug builds w/ symbols (default=optimized).\n  -a, --asan            Download AddressSanitizer builds.\n  -t, --tsan            Download ThreadSanitizer builds.\n  --fuzzing             Download --enable-fuzzing builds.\n  --coverage            Download --coverage builds. This also pulls down the\n                        *.gcno files\n  --valgrind            Download Valgrind builds.\n\nTest Arguments:\n  --tests  [ ...]       Download tests associated with this build. Acceptable\n                        values are: gtest, common, reftests\n  --full-symbols        Download the full crashreport-symbols.zip archive.\n\nMisc. Arguments:\n  -n NAME, --name NAME  Specify a name (default=auto)\n  -o OUT, --out OUT     Specify output directory (default=.)\n  --dry-run             Search for build and output metadata only, don't\n                        download anything.\n\nNear Arguments:\n  If the specified build isn't found, iterate over builds in the specified\n  direction\n\n  --nearest-newer       Search from specified build in ascending order\n  --nearest-older       Search from the specified build in descending order\n\nBoundary Arguments (YYYY-MM-DD or SHA1 revision):\n  --start START         Start build id (default: earliest available build)\n  --end END             End build id (default: latest available build)\n\nBisection Arguments:\n  --timeout TIMEOUT     Maximum iteration time in seconds (default: 60)\n  --repeat REPEAT       Number of times to evaluate testcase (per build)\n  --config CONFIG       Path to optional config file\n  --find-fix            Identify fix date\n\n```\n\nSimple Bisection\n----------------\n```\npython -m autobisect firefox trigger.html --prefs prefs.js --asan --end 2017-11-14\n```\n\nBy default, Autobisect will cache downloaded builds (up to 30GBs) to reduce bisection time.  This behavior can be modified by supplying a custom configuration file in the following format:\n```\n[autobisect]\nstorage-path: /home/ubuntu/cached\npersist: true\n; size in MBs\npersist-limit: 30000\n```\n\nDevelopment\n-----------\nAutobisect includes a pre-commit hook for [black](https://github.com/psf/black) and [flake8](https://flake8.pycqa.org/en/latest/).  To install the pre-commit hook, run the following.  \n```bash\npre-commit install\n```\n\nFurthermore, all tests should be executed via tox.\n```bash\npoetry run tox\n```\n\n",
    'author': 'Jason Kratzer',
    'author_email': 'jkratzer@mozilla.com',
    'maintainer': 'Mozilla Fuzzing Team',
    'maintainer_email': 'fuzzing@mozilla.com',
    'url': 'https://github.com/MozillaSecurity/autobisect',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from grizzly.replay import ReplayArgs

from ...args import BisectCommonArgs, _remove_arg, _suppress_arg


class BrowserArgs(BisectCommonArgs, ReplayArgs):
    """
    Argparser for BrowserEvaluator
    """

    CONFLICTING_ARGS = [
        "binary",
        "fuzzmanager",
        "include_test",
        "input",
        "logs",
        "no_harness",
        "relaunch",
        "rr",
        "tool",
    ]

    IGNORED_ARGS = [
        "min_crashes",
        "platform",
        "sig",
        "working_path",
    ]

    def __init__(self, parser):
        self.parser = parser
        super().__init__()

        # Remove conflicting args
        for arg in BrowserArgs.CONFLICTING_ARGS:
            self._sanity_skip.add(arg)
            _remove_arg(self.parser, arg)

        # Suppress help output for unused args
        for arg in BrowserArgs.IGNORED_ARGS:
            self._sanity_skip.add(arg)
            _suppress_arg(self.parser, arg)

    def parse_args(self, argv=None):
        args = self.parser.parse_args(argv)
        self.sanity_check(args)
        return args



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
import re
from datetime import datetime, timedelta

import pytest
from freezegun import freeze_time
from fuzzfetch import BuildFlags, Fetcher, Platform

from autobisect import EvaluatorResult
from autobisect.bisect import (
    Bisector,
    StatusException,
    VerificationStatus,
    get_autoland_range,
)
from autobisect.builds import BuildRange


class MockFetcher:
    """
    Class for mocking Fetcher objects
    """

    def __init__(self, dt=None, changeset=None):
        self.datetime = dt
        self.changeset = changeset


class MockBisector(Bisector):
    """
    Class for mocking Bisector objects
    """

    # pylint: disable=super-init-not-called
    def __init__(self, start, end):
        self.start = MockFetcher(dt=start)
        self.end = MockFetcher(dt=end)
        self.target = "firefox"
        self.branch = "central"
        self.find_fix = False
        self.flags = BuildFlags(
            asan=False,
            tsan=False,
            debug=False,
            fuzzing=False,
            coverage=False,
            valgrind=False,
        )
        self.platform = Platform("Linux", "x86_64")


def test_bisect_get_daily_builds_simple():
    """
    Simple test for Bisector._get_daily_builds
    """
    start_date = datetime.now() - timedelta(days=10 + 1)
    end_date = datetime.now()

    bisector = MockBisector(start_date, end_date)
    builds = bisector._get_daily_builds()

    assert isinstance(builds, BuildRange)  # Returns BuildRange
    assert all(re.match(r"\d{4}-\d{2}-\d{2}", b) is not None for b in builds)
    assert len(builds) == 10  # Length 365 - 1


def test_bisect_get_daily_builds_no_builds():
    """
    Test Bisector._get_daily_builds using an empty build range
    """
    start_date = datetime.now()
    bisector = MockBisector(start_date, start_date)
    builds = bisector._get_daily_builds()

    assert isinstance(builds, BuildRange)  # Returns BuildRange
    assert len(builds) == 0  # Length 365 - 1


def test_bisect_get_pushdate_builds_simple(mocker):
    """
    Simple test of Bisector._get_pushdate_builds
    """
    end_date = datetime(2020, 1, 1, 0, 0)
    start_date = end_date - timedelta(days=1)
    bisector = MockBisector(start_date, end_date)

    # Generate fake pushdates
    pushdates = []
    for dt in [start_date, end_date]:
        group = []
        end_of_day = dt.replace(hour=23, minute=59, second=59)
        delta = (end_of_day - dt) / 10
        current = dt
        while current < end_of_day:
            group.append(MockFetcher(dt=current))
            current = current + delta
        pushdates.append(group)

    mocker.patch("autobisect.bisect.Fetcher.iterall", side_effect=pushdates)
    builds = bisector._get_pushdate_builds()

    assert isinstance(builds, BuildRange)
    assert len(builds) == 9


@pytest.mark.usefixtures("requests_mock_cache")
def test_bisect_get_autoland_builds_simple():
    """
    Simple test of Bisector._get_autoland_builds
    """
    bisector = MockBisector(None, None)
    bisector.start = MockFetcher(dt=datetime(2019, 12, 30), changeset="03ed5ed6cba7")
    bisector.end = MockFetcher(dt=datetime(2019, 12, 31), changeset="a1266665b89b")
    builds = bisector._get_autoland_builds()
    repo_url = "https://hg.mozilla.org/integration/autoland"

    assert isinstance(builds, BuildRange)
    assert len(builds) > 1
    assert all(b.build_info["moz_source_repo"] == repo_url for b in builds)


@pytest.mark.usefixtures("requests_mock_cache")
def test_get_autoland_range_simple():
    """
    Simple test of get_autoland_range()
    """
    with freeze_time("2020-01-01"):
        changesets = get_autoland_range("03ed5ed6cba7", "a1266665b89b")

        assert len(changesets) == 28
        for changeset in changesets:
            assert isinstance(changeset, str) and len(changeset) == 40


@pytest.mark.usefixtures("requests_mock_cache")
def test_get_autoland_range_invalid_revs():
    """
    Test get_autoland_range using invalid revisions
    """
    assert get_autoland_range("foo", "bar") is None


@pytest.mark.usefixtures("requests_mock_cache")
def test_get_autoland_range_multiple_revs():
    """
    Test that get_autoland_range returns None when multiple changsets identified
    """
    assert get_autoland_range("385f49adaf00", "590613078c74") is None


@pytest.mark.parametrize("status", EvaluatorResult)
@pytest.mark.parametrize("find_fix", [True, False])
def test_update_range_simple(status, find_fix):
    """
    Simple test of Bisector.update_range
    """
    builds = []
    for _ in range(10):
        next_date = datetime(2020, 1, 1, 0, 0) + timedelta(days=1)
        builds.append(MockFetcher(dt=next_date))

    for index, build in enumerate(builds):
        bisector = MockBisector(builds[0], builds[-1])
        bisector.find_fix = find_fix
        build_range = BuildRange(builds)
        bisector.update_range(status, build, index, build_range)
        if status == EvaluatorResult.BUILD_PASSED:
            if find_fix:
                assert bisector.end == build
            else:
                assert bisector.start == build
        elif status == EvaluatorResult.BUILD_CRASHED:
            if find_fix:
                assert bisector.start == build
            else:
                assert bisector.end == build


@pytest.mark.parametrize("find_fix", [True, False])
@pytest.mark.parametrize("end_result", EvaluatorResult)
@pytest.mark.parametrize("start_result", EvaluatorResult)
# pylint: disable=inconsistent-return-statements
def test_verify_bounds_simple(mocker, start_result, end_result, find_fix):
    """
    Verify expected verify bounds status
    """
    bisector = MockBisector(None, None)
    bisector.find_fix = find_fix

    mocker.patch(
        "autobisect.bisect.Bisector.test_build", side_effect=[start_result, end_result]
    )
    result = bisector.verify_bounds()

    if start_result == EvaluatorResult.BUILD_FAILED:
        assert result == VerificationStatus.START_BUILD_FAILED
    elif start_result == EvaluatorResult.BUILD_CRASHED and not find_fix:
        assert result == VerificationStatus.START_BUILD_CRASHES
    elif start_result == EvaluatorResult.BUILD_PASSED and find_fix:
        assert result == VerificationStatus.FIND_FIX_START_BUILD_PASSES
    elif end_result == EvaluatorResult.BUILD_FAILED:
        assert result == VerificationStatus.END_BUILD_FAILED
    elif end_result == EvaluatorResult.BUILD_PASSED and not find_fix:
        return VerificationStatus.END_BUILD_PASSES
    elif end_result == EvaluatorResult.BUILD_CRASHED and find_fix:
        return VerificationStatus.FIND_FIX_END_BUILD_CRASHES
    else:
        assert result == VerificationStatus.SUCCESS


@pytest.mark.parametrize(
    "test_results",
    [[EvaluatorResult.BUILD_PASSED, None], [None, EvaluatorResult.BUILD_CRASHED]],
)
def test_verify_bounds_invalid_status(mocker, test_results):
    """
    Verify that invalid test results throw a StatusException
    """
    bisector = MockBisector(None, None)
    mocker.patch("autobisect.bisect.Bisector.test_build", side_effect=test_results)
    with pytest.raises(StatusException):
        bisector.verify_bounds()


def test_build_iterator_random(mocker):
    """
    Verify that random.choice called when random_choice arg set to True
    """
    builds = []
    for _ in range(1, 4):
        builds.append(mocker.Mock(spec=Fetcher))
    spy = mocker.patch("autobisect.bisect.random.choice", side_effect=builds)

    bisector = MockBisector(None, None)
    generator = bisector.build_iterator(builds, True)
    next_build = next(generator)
    while next_build is not None:
        next_build = generator.send(EvaluatorResult.BUILD_PASSED)

    assert spy.call_count == 3



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import argparse
import logging
import os
import tempfile

from grizzly.common import TestCase
from grizzly.replay import ReplayArgs, ReplayManager

from ..base import Evaluator, EvaluatorResult

LOG = logging.getLogger("browser-eval")

# Disable sub loggers
logging.getLogger("grizzly.replay").setLevel(logging.WARNING)


class ArgParserNoExit(argparse.ArgumentParser):
    """ Override default ArgParser SystemExit Behavior """

    def exit(self, status=0, message=None):
        pass

    def error(self, message):
        raise BrowserEvaluatorException(message)


class ReplayArgsNoExit(ReplayArgs):
    """ Set parser to ArgParserNoExit instance """

    def __init__(self, *args, **kwds):
        self.parser = ArgParserNoExit()
        super().__init__(*args, **kwds)


class BrowserEvaluatorException(Exception):
    """ Simple exception handler for BrowserEvaluator """


class BrowserEvaluator(Evaluator):
    """
    Testcase evaluator for Firefox
    """

    def __init__(self, testcase, **kwargs):
        self.testcase = testcase

        # FFPuppet arguments
        self._repeat = kwargs.get("repeat", 1)
        self._ignore = kwargs.get("ignore", ["log-limit", "memory", "timeout"])
        self._use_valgrind = kwargs.get("valgrind", False)
        self._use_xvfb = kwargs.get("xvfb", True)
        self._timeout = kwargs.get("timeout", 60)
        self._launch_timeout = kwargs.get("launch_timeout", 300)
        self._prefs = kwargs.get("prefs", None)
        self._env_vars = kwargs.get("env", None)

        if logging.getLogger().level != logging.DEBUG:
            logging.getLogger("grizzly").setLevel(logging.WARNING)

    def verify_build(self, binary):
        """
        Verify that build doesn't crash on start
        :param binary: The path to the target binary
        :return: Boolean
        """
        with tempfile.NamedTemporaryFile(suffix=".html", mode="w") as temp:
            temp.write("<html><script>window.close()</script></html>")
            temp.flush()
            LOG.info("> Verifying build...")

            status = self.launch(binary, temp.name, 1)

        if status != EvaluatorResult.BUILD_PASSED:
            LOG.error(">> Build crashed!")
            return False

        LOG.info(">> Build verified!")
        return True

    def evaluate_testcase(self, build_path):
        """
        Validate build and launch with supplied testcase
        :return: Result of evaluation
        """
        binary = os.path.join(build_path, "firefox")
        result = EvaluatorResult.BUILD_FAILED
        if os.path.isfile(binary) and self.verify_build(binary):
            LOG.info("> Launching build with testcase...")
            result = self.launch(binary, self.testcase, self._repeat, scan_dir=True)

        if result == EvaluatorResult.BUILD_CRASHED:
            LOG.info(">> Build crashed!")
        else:
            LOG.info(">> Build did not crash!")

        return result

    def launch(self, binary, test_path, repeat, scan_dir=False):
        """
        Launch firefox using the supplied binary and testcase
        :param binary: The path to the firefox binary
        :param test_path: The path to the testcase
        :param repeat: The number of times to launch the browser
        :param scan_dir: Scan subdirectory for additional files to serve
        :return: The return code or None
        """
        # Create testcase
        testcase = TestCase.load_single(test_path, load_prefs=True, adjacent=scan_dir)
        if self._env_vars:
            for key, value in self._env_vars.items():
                testcase.add_environ_var(key, value)

        with tempfile.TemporaryDirectory() as test_dir:
            testcase.dump(test_dir, include_details=True)

            raw_args = [
                binary,
                test_dir,
                "--timeout",
                self._timeout,
                "--launch-timeout",
                self._launch_timeout,
                "--repeat",
                repeat,
                "--relaunch",
                1,
                "--no-harness",
                "--ignore",
                *self._ignore,
            ]

            if self._prefs:
                raw_args.extend(["--prefs", self._prefs])
            if self._use_valgrind:
                raw_args.append("--valgrind")
            if self._use_xvfb:
                raw_args.append("--xvfb")

            args = ReplayArgsNoExit().parse_args([str(arg) for arg in raw_args])
            success = ReplayManager.main(args)

            if not success:
                return EvaluatorResult.BUILD_CRASHED

            return EvaluatorResult.BUILD_PASSED
